// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include <sys/eventfd.h>	// for eventfd, EFD_SEMAPHORE

#include <cerrno>		// for ECANCELED, ETIME
#include <cstdint>	// for uint64_t, int32_t

#include "Common.hpp"
#include "Emper.hpp"
#include "fixtures/assert.hpp"
#include "io/Future.hpp"	// for ReadFuture, WriteFuture

using emper::io::ReadFuture;
using emper::io::WriteFuture;

int efd;
uint64_t read_buf;
uint64_t write_buf = 42;

void cancelSubmittedCallback() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	readFuture.setCallback([](int32_t res) { ASSERT(res == -ECANCELED) });
	readFuture.submit();
	int32_t res = readFuture.cancel();

	if constexpr (emper::IO_SINGLE_URING) {
		// When using a single io_uring it is possible the submitted Future was
		// canceled before reaching the io_uring.
		ASSERT(!res || res == -ECANCELED);
	} else {
		ASSERT(!res);
	}
}

void cancelNotSubmitted() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	ASSERT(readFuture.cancel() == -ENOENT);
}

void cancelSubmittedNotCompleted() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	readFuture.submit();
	ASSERT(readFuture.cancel() == -ECANCELED);
}

void cancelCompleted() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	readFuture.submit();
	WriteFuture writeFuture(efd, &write_buf, sizeof(write_buf), 0);
	ASSERT(writeFuture.submitAndWait() == sizeof(write_buf));
	ASSERT(readFuture.cancel() == sizeof(write_buf) && read_buf == write_buf);
}

void cancelNotCompletedChain() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	ReadFuture readFuture2(efd, &read_buf, sizeof(read_buf), 0);
	readFuture2.setDependency(readFuture);

	readFuture2.submit();
	ASSERT(readFuture2.cancel() == -ECANCELED);
	ASSERT(readFuture.wait() == -ECANCELED);
}

void cancelPartialCompletedChain() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	ReadFuture readFuture2(efd, &read_buf, sizeof(read_buf), 0);
	readFuture2.setDependency(readFuture);

	readFuture2.submit();
	ASSERT(readFuture2.cancel() == -ECANCELED);
	ASSERT(readFuture.wait() == -ECANCELED);
}

void cancelNotCompletedFutureChain() {
	ReadFuture readFuture(efd, &read_buf, sizeof(read_buf), 0);
	WriteFuture writeFuture(efd, &write_buf, sizeof(write_buf), 0);
	ReadFuture readFuture2(efd, &read_buf, sizeof(read_buf), 0);
	readFuture2.setDependency(readFuture);

	readFuture2.submit();
	ASSERT(writeFuture.submitAndWait() == sizeof(write_buf));
	// TODO: investigate why this read is completed with -EINTR most of the time
	int r = readFuture2.cancel();
	ASSERT(r == -EINTR || r == -ECANCELED);
	ASSERT(readFuture.wait() == sizeof(write_buf) && read_buf == write_buf);
}

void emperTest() {
	efd = eventfd(0, 0);
	if (efd == -1) {
		DIE_MSG_ERRNO("eventfd failed");
	}

	cancelSubmittedCallback();
	cancelNotSubmitted();
	cancelSubmittedNotCompleted();
	cancelCompleted();
	cancelNotCompletedChain();
	cancelPartialCompletedChain();
}
