// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include <bits/types/struct_iovec.h>	// for iovec
#include <fcntl.h>										// for open, O_RDONLY
#include <sys/types.h>

#include <array>
#include <cstdlib>	// for mkstemp
#include <cstring>	// for memcmp, memset
#include <memory>		// for allocator, unique_ptr
#include <string>		// for string
#include <vector>		// for vector

#include "Common.hpp"
#include "Debug.hpp"
#include "Runtime.hpp"
#include "fixtures/assert.hpp"
#include "fixtures/network.hpp"
#include "io.hpp"
#include "io/Future.hpp"
class Fiber;

using namespace emper::io;

static const std::string HOST = "127.0.0.1";
static const std::string PORT = "4243";
#define MAX 1024

static void server_func(int client_fd) {
	std::array<char, MAX> recv_buf;
	std::array<char, MAX> read_buf;

	for (;;) {
		ssize_t received = recvAndWait(client_fd, recv_buf.data(), recv_buf.size(), 0);
		if (received == 0) {
			return;
		}

		if (received == -1) {
			DIE_MSG_ERRNO("recv failed");
			break;
		}

		DBG("Server got msg from client -> writing to file");

		// NOLINTNEXTLINE(modernize-avoid-c-arrays)
		char file_name[] = "/tmp/emper-SimpleDiskAndNetworkTestFile-XXXXXX";
		int file_fd = mkstemp(file_name);
		if (file_fd == -1) {
			DIE_MSG_ERRNO("mkstemp failed");
		}

		ssize_t written = writeFileAndWait(file_fd, recv_buf.data(), received);
		if (written < 0) {
			DIE_MSG_ERRNO("write failed");
		}
		closeAndWait(file_fd);

		DBG("Msg written to file");

		file_fd = openAndWait(file_name, O_RDONLY);
		if (file_fd == -1) {
			DIE_MSG_ERRNO("open failed");
		}

		ssize_t bytes_read = readFileAndWait(file_fd, read_buf.data(), written);
		if (bytes_read == 0) {
			DIE_MSG("nothing to read");
		}

		if (bytes_read < 0) {
			DIE_MSG_ERRNO("read failed");
		}
		closeAndWait(file_fd);

		DBG("Msg read from file");

		ssize_t sent = sendAndWait(client_fd, read_buf.data(), bytes_read, 0);
		if (sent == 0) {
			DIE_MSG("client socket unexpected shutdown");
		}

		if (sent == -1) {
			DIE_MSG_ERRNO("send failed");
		}

		DBG("Msg sent back to client");
	}
}

static void testDiskAndNetwork() {
	Fiber* listener = tcp_listener(HOST, PORT, server_func, 1024, {SockOpt::ReusePort});
	if (!listener) {
		DIE_MSG("creating listener failed");
	}
	Runtime::getRuntime()->schedule(*listener);

	const std::vector<std::string> msgs{"foo", "bar"};
	echo_client(HOST, PORT, msgs);
}

static void testIov() {
	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	char file_name[] = "/tmp/emper-SimpleDiskAndNetworkTestFile-XXXXXX";
	int file_fd = mkstemp(file_name);
	if (file_fd == -1) {
		DIE_MSG_ERRNO("mkstemp failed");
	}

	DBG("Opened temp file to test writev");

	std::array<char, MAX> read_buf;

	const int iovcnt = 2;
	std::array<struct iovec, iovcnt> iov;

	std::string s1 = "foo";
	std::string s2 = "bar";

	iov[0].iov_base = (void*)s1.c_str();
	iov[0].iov_len = s1.length();
	iov[1].iov_base = (void*)s2.c_str();
	iov[1].iov_len = s2.length();

	DBG("Prepared iov");

	auto writevFuture = emper::io::writev(file_fd, &iov[0], iovcnt);
	ssize_t written = writevFuture->waitAndSetErrno();
	if (written < 0) {
		DIE_MSG_ERRNO("writev failed");
	}
	closeAndWait(file_fd);

	file_fd = openAndWait(file_name, O_RDONLY);
	if (file_fd == -1) {
		DIE_MSG_ERRNO("open failed");
	}

	DBG("Reread written data");

	auto readFuture = readFile(file_fd, read_buf.data(), written, 0, true);
	ssize_t bytes_read = readFuture->waitAndSetErrno();
	if (bytes_read == 0) {
		DIE_MSG("nothing to read");
	}

	if (bytes_read < 0) {
		DIE_MSG_ERRNO("read failed");
	}

	ASSERT(written == bytes_read);

	ASSERT(memcmp(read_buf.data(), iov[0].iov_base, iov[0].iov_len) == 0);
	ASSERT(memcmp(read_buf.data() + iov[0].iov_len, iov[1].iov_base, iov[1].iov_len) == 0);

	closeAndWait(file_fd);
}

void emperTest() {
	testDiskAndNetwork();
	testIov();
}
